@@ -1,12 +1,13 @@
 package SVK::Command::Copy;
 use strict;
-our $VERSION = $SVK::VERSION;
+use SVK::Version;  our $VERSION = $SVK::VERSION;
 use base qw( SVK::Command::Mkdir );
-use SVK::Util qw( get_anchor get_prompt abs2rel );
+use SVK::Util qw( get_anchor get_prompt abs2rel splitdir is_uri make_path );
 use SVK::I18N;
 
 sub options {
     ($_[0]->SUPER::options,
+     'q|quiet'         => 'quiet',
      'r|revision=i' => 'rev');
 }
 
@@ -17,19 +18,44 @@ sub parse_arg {
     push @arg, '' if @arg == 1;
 
     my $dst = pop(@arg);
-    my @src = (map {$self->arg_co_maybe ($_)} @arg);
+    die loc ("Copy destination can't be URI.\n")
+	if is_uri ($dst);
 
-    if ( my $target = eval { $self->arg_co_maybe ($dst) }) {
+    die loc ("More than one URI found.\n")
+	if (grep {is_uri($_)} @arg) > 1;
+    my @src;
+
+    if ( my $target = eval { $self->{xd}->target_from_copath_maybe($dst) }) {
         $dst = $target;
+	# don't allow new uri in source when target is copath
+	@src = (map {$self->arg_co_maybe
+			 ($_, $dst->{copath}
+			  ? loc ("path '%1' is already a checkout", $dst->{report})
+			  : undef)} @arg);
     }
     else {
-        $self->{_checkout_path} = $dst;
-        my $path = get_prompt(loc("Enter a depot path to copy into (under // if no leading '/'): "));
-        $path =~ s{^//+}{};
-        $path =~ s{//+}{/};
-        $path = "//$path" unless $path =~ m!^/!;
-        $path = "$path/" unless $path =~ m!/\z!;
-        $dst = $self->arg_depotpath($path);
+	@src = (map {$self->arg_co_maybe ($_)} @arg);
+        # Asking the user for copy destination.
+        # In this case, first magically promote ourselves to "cp -p".
+        # (otherwise it hurts when user types //deep/directory/name)
+        $self->{parent} = 1;
+
+        # -- make a sane default here for mirroring --
+        my $default = undef;
+        if (@src == 1 and $src[0]->path =~ m{/mirror/([^/]+)$}) {
+            $default = "/" . $src[0]->depotname . "/$1";
+        }
+
+        my $path = $self->prompt_depotpath("copy", $default);
+
+        if ($dst eq '.') {
+            $self->{_checkout_path} = (splitdir($path))[-1];
+        }
+        else {
+            $self->{_checkout_path} = $dst;
+        }
+
+        $dst = $self->arg_depotpath("$path/");
     }
 
     return (@src, $dst);
@@ -37,21 +63,7 @@ sub parse_arg {
 
 sub lock {
     my $self = shift;
-    $_[-1]->{copath} ? $self->lock_target ($_[-1]) : $self->lock_none;
-}
-
-sub do_copy_direct {
-    # OBSOLETED
-    my ($self, %arg) = @_;
-    my $fs = $arg{repos}->fs;
-    my $edit = $self->get_commit_editor ($fs->revision_root ($fs->youngest_rev),
-					 sub { print loc("Committed revision %1.\n", $_[0]) },
-					 '/', %arg);
-    # XXX: check parent, check isfile, check everything...
-    $edit->open_root();
-    $edit->copy_directory ($arg{dpath}, "file://$arg{repospath}$arg{path}",
-			   $arg{revision});
-    $edit->close_edit();
+    $self->lock_coroot($_[-1]);
 }
 
 sub handle_co_item {
@@ -60,29 +72,50 @@ sub handle_co_item {
     my $xdroot = $dst->root ($self->{xd});
     die loc ("Path %1 does not exist.\n", $src->{path})
 	if $src->root->check_path ($src->{path}) == $SVN::Node::none;
-    die loc ("Path %1 already exists.\n", $dst->{copath})
-	if -e $dst->{copath};
     my ($copath, $report) = @{$dst}{qw/copath report/};
-    $src->normalize;
-    $src->anchorify; $dst->anchorify;
+    die loc ("Path %1 already exists.\n", $copath)
+	if -e $copath;
+    my ($entry, $schedule) = $self->{xd}->get_entry($copath);
+    $src->normalize; $src->anchorify;
+    $self->ensure_parent($dst);
+    $dst->anchorify;
+
+    my $notify = $self->{quiet} ? SVK::Notify->new(quiet => 1) : undef;
     # if SVK::Merge could take src being copath to do checkout_delta
     # then we have 'svk cp copath... copath' for free.
+    # XXX: use editor::file when svkup branch is merged
     SVK::Merge->new (%$self, repos => $dst->{repos}, nodelay => 1,
-		     report => $report,
+		     report => $report, notify => $notify,
 		     base => $src->new (path => '/', revision => 0),
-		     src => $src, dst => $dst)->run ($self->get_editor ($dst));
+		     src => $src, dst => $dst)->run
+			 ($self->{xd}->get_editor
+			  ( %$dst,
+			    ignore_checksum => 1,
+			    targets => undef,
+			    quiet => 1,
+			    oldroot => $dst->root($self->{xd}),
+			    newroot => $dst->root($self->{xd}),
+			    target => $dst->{targets}[0] || '',
+			    update => 1,
+			    ignore_keywords => 1,
+			    check_only => $self->{check_only}));
 
-    $self->{xd}{checkout}->store_recursively ($copath, {'.schedule' => undef,
-							'.newprop' => undef});
-    $self->{xd}{checkout}->store ($copath, {'.schedule' => 'add',
+    $self->{xd}{checkout}->store_recursively
+	($copath, { revision => undef });
+    # XXX: can the scheudle be something other than delete ?
+    $self->{xd}{checkout}->store ($copath, {'.schedule' => $schedule ? 'replace' : 'add',
 					    scheduleanchor => $copath,
 					    '.copyfrom' => $src->path,
 					    '.copyfrom_rev' => $src->{revision}});
 }
 
 sub handle_direct_item {
-    my ($self, $editor, $anchor, $m, $src, $dst) = @_;
+    my ($self, $editor, $anchor, $m, $src, $dst, $other_call) = @_;
     $src->normalize;
+    # if we have targets, ->{path} must exist
+    if (!$self->{parent} && $dst->{targets} && !$dst->root->check_path ($dst->{path})) {
+	die loc ("Parent directory %1 doesn't exist, use -p.\n", $dst->{report});
+    }
     my ($path, $rev) = @{$src}{qw/path revision/};
     if ($m) {
 	$path =~ s/^\Q$m->{target_path}\E/$m->{source}/;
@@ -92,14 +125,16 @@ sub handle_direct_item {
     else {
 	$path = "file://$src->{repospath}$path";
     }
-    $editor->close_directory
-	($editor->add_directory (abs2rel ($dst->path, $anchor => undef, '/'), 0, $path, $rev));
+    my $baton = $editor->add_directory (abs2rel ($dst->path, $anchor => undef, '/'), 0, $path, $rev);
+    $other_call->($baton) if $other_call;
+    $editor->close_directory($baton);
     $self->adjust_anchor ($editor);
 }
 
 sub _unmodified {
     my ($self, $target) = @_;
-    $target->anchorify;
+    # Use condensed to do proper anchorification.
+    $target = $self->arg_condensed ($target->copath);
     $self->{xd}->checkout_delta
 	( %$target,
 	  xdroot => $target->root ($self->{xd}),
@@ -108,8 +143,9 @@ sub _unmodified {
 	    ( cb_flush => sub {
 		  die loc ("%1 is modified.\n", $target->copath ($_[0]));
 	      })),
+	  # need tests: only useful for move killing the src with unknown entries
 	  cb_unknown => sub {
-	      die loc ("%1 is missing.\n", $target->copath ($_[0]))});
+	      die loc ("%1 is unknown.\n", $target->copath ($_[1]))});
 }
 
 sub check_src {
@@ -133,9 +169,12 @@ sub run {
     # XXX: check dst to see if the copy is obstructured or missing parent
     my $fs = $dst->{repos}->fs;
     if ($dst->{copath}) {
-	# XXX: check if dst is versioned
-	return loc("%1 is not a directory.\n", $dst->{copath})
+	return loc("%1 is not a directory.\n", $dst->{report})
 	    if $#src > 0 && !-d $dst->{copath};
+	return loc("%1 is not a versioned directory.\n", $dst->{report})
+	    if -d $dst->{copath} &&
+		!($dst->root($self->{xd})->check_path ($dst->path) ||
+		  $self->{xd}{checkout}->get ($dst->{copath})->{'.schedule'});
 	my @cpdst;
 	for (@src) {
 	    my $cpdst = $dst->new;
@@ -164,7 +203,7 @@ sub run {
 	$self->finalize_dynamic_editor ($editor);
     }
 
-    if (my $copath = $self->{_checkout_path}) {
+    if (defined( my $copath = $self->{_checkout_path} )) {
         my $checkout = $self->command ('checkout');
 	$checkout->getopt ([]);
         my @arg = $checkout->parse_arg ($dst->{report}, $copath);
@@ -186,16 +225,20 @@ SVK::Command::Copy - Make a versioned copy
 =head1 SYNOPSIS
 
  copy DEPOTPATH1 DEPOTPATH2
- copy DEPOTPATH PATH
+ copy DEPOTPATH [PATH]
 
 =head1 OPTIONS
 
- -r [--revision] arg    : act on revision ARG instead of the head revision
- -m [--message] arg     : specify commit message ARG
+ -r [--revision] REV    : act on revision REV instead of the head revision
  -p [--parent]          : create intermediate directories as required
- -P [--patch] arg       : instead of commit, save this change as a patch
- -C [--check-only]      : try operation but make no changes
+ -m [--message] MESSAGE : specify commit message MESSAGE
+ -F [--file] FILENAME   : read commit message from FILENAME
+ --template             : use the specified message as the template to edit
+ --encoding ENC         : treat -m/-F value as being in charset encoding ENC
+ -P [--patch] NAME      : instead of commit, save this change as a patch
  -S [--sign]            : sign this change
+ -C [--check-only]      : try operation but make no changes
+ --direct               : commit directly even if the path is mirrored
 
 =head1 AUTHORS
 
@@ -203,7 +246,7 @@ Chia-liang Kao E<lt>clkao@clkao.orgE<gt>
 
 =head1 COPYRIGHT
 
-Copyright 2003-2004 by Chia-liang Kao E<lt>clkao@clkao.orgE<gt>.
+Copyright 2003-2005 by Chia-liang Kao E<lt>clkao@clkao.orgE<gt>.
 
 This program is free software; you can redistribute it and/or modify it
 under the same terms as Perl itself.
